import java.util.concurrent.Semaphore; // Semaphores library for Java

class ProducerConsumerMulti {
	final int CAPACITY = 10;
	final int CONSUMERS = 2;
	final int PRODUCERS = 3;
	
    private ModCounter write = new ModCounter(CAPACITY);
    private ModCounter read = new ModCounter(CAPACITY);
	
    int[] buffer = new int[CAPACITY];
    Semaphore empty = new Semaphore(CAPACITY);
    Semaphore full = new Semaphore(0);
    
    public ProducerConsumerMulti() {
    	Thread[] p = new Thread[PRODUCERS];
    	Thread[] c = new Thread[CONSUMERS];
    	
    	for (int i=0 ; i<PRODUCERS ; ++i) {
    		p[i] = new Thread(new Producer(i));
    	}
    	for (int i=0 ; i<CONSUMERS ; ++i) {
    		c[i] = new Thread(new Consumer(i));
    	}

    	for (int i=0 ; i<CAPACITY ; ++i) {
    		buffer[i] = 0;
    	}
    	
    	for (int i=0 ; i<PRODUCERS ; ++i) {
    		p[i].start();
    	}
    	for (int i=0 ; i<CONSUMERS ; ++i) {
    		c[i].start();
    	}
    }

    public static void main(String[] args) {
        new ProducerConsumerMulti();
    }

    class Producer implements Runnable {        
        private int id;
        
        Producer(int i) {
        	id = i;
        }
    	
        public void run() {
        	while (true) {
        		synchronized(empty) {
        			while (!empty.tryAcquire()) {
        				try {
        					empty.wait();
        				} catch (InterruptedException e) {
        					System.out.print(e.getStackTrace());
        				}
        			}
        		}
        		synchronized (write) {
        			buffer[write.value()] = 1;
        			System.out.println("Produced " + write.value() + "(" + id + ")");
        			write.inc();
        		}

        		synchronized(full) {
        			full.release();
        			full.notify();
        		}
        	}
        }
    }

    class Consumer implements Runnable {
        private int id;
        
        Consumer(int i) {
        	id = i;
        }

        public void run() {
        	while (true) {
        		// Notice that the order between synchronization
        		// and the acquisition is slightly different.
        		// The synchronized block is larger here 
        		// than what we saw in class
        		// The code here is more succinct and clearer
        		synchronized(full) {
        			while (!full.tryAcquire()) {
        				try {
        					full.wait();
        				} catch (InterruptedException e) {
        					System.out.print(e.getStackTrace());
        				}
        			}
        		}
        		
        		// I added a structure wrapping the modulo counter
        		// this way we can synchronize on this object
        		// Notice that here internal synchronization would have 
        		// to work differently.
        		// For example, by having the inc method return the
        		// previous value of the counter, storing this value
        		// and then using the stored local value to change the 
        		// buffer and to print.
        		// However, the printouts would very likely be
        		// out of order!
        		synchronized (read) {
        			buffer[read.value()] = 0;
        			System.out.println("\t\t\tConsumed " + read.value() + "(" + id + ")");
        			read.inc();
        		}

        		synchronized(empty) {
        			empty.release();
        			empty.notify();
        		}
        	}
        }
    }

    class ModCounter {
    	private int count;
    	private final int modulo;
    	
    	ModCounter(int m) {
    		modulo = m;
    		count = 0;
    	}
    	
    	public void inc() {
    		count = (count+1)%modulo;
    	}
    	
    	public int value() {
    		return count;
    	}
    }


}
